@@ -13,7 +13,9 @@ The package resource API is designed to work with normal filesystem packages,
 method.
 """
 
-import sys, os, zipimport, time, re, imp, types
+import sys, os, time, re, imp, types, zipfile, zipimport
+import warnings
+import stat
 from urlparse import urlparse, urlunparse
 
 try:
@@ -987,6 +989,7 @@ variable to point to an accessible directory.
         extract, as it tracks the generated names for possible cleanup later.
         """
         extract_path = self.extraction_path or get_default_cache()
+        self._warn_unsafe_extraction_path(extract_path)
         target_path = os.path.join(extract_path, archive_name+'-tmp', *names)
         try:
             _bypass_ensure_directory(target_path)
@@ -996,6 +999,29 @@ variable to point to an accessible directory.
         self.cached_files[target_path] = 1
         return target_path
 
+    @staticmethod
+    def warn_unsafe_extraction_path(path):
+        """
+        If the default extraction path is overridden and set to an insecure
+        location, such as /tmp, it opens up an opportunity for an attacker to
+        replace an extracted file with an unauthorized payload. Warn the user
+        if a known insecure location is used.
+
+        See Distribute #375 for more details.
+        """
+        if os.name == 'nt' and not path.startswith(os.environ['windir']):
+            # On Windows, permissions are generally restrictive by default
+            #  and temp directories are not writable by other users, so
+            #  bypass the warning.
+            return
+        mode = os.stat(path).st_mode
+        if mode & stat.S_IWOTH or mode & stat.S_IWGRP:
+            msg = ("%s is writable by group/others and vulnerable to attack "
+                "when "
+                "used with get_resource_filename. Consider a more secure "
+                "location (set with .set_extraction_path or the "
+                "PYTHON_EGG_CACHE environment variable)." % path)
+            warnings.warn(msg, UserWarning)
 
 
 
@@ -1349,6 +1375,37 @@ class EmptyProvider(NullProvider):
 empty_provider = EmptyProvider()
 
 
+def build_zipmanifest(path):
+    """
+    This builds a similar dictionary to the zipimport directory
+    caches.  However instead of tuples, ZipInfo objects are stored.
+
+    The translation of the tuple is as follows:
+      * [0] - zipinfo.filename on stock pythons this needs "/" --> os.sep
+              on pypy it is the same (one reason why distribute did work
+              in some cases on pypy and win32).
+      * [1] - zipinfo.compress_type
+      * [2] - zipinfo.compress_size
+      * [3] - zipinfo.file_size
+      * [4] - len(utf-8 encoding of filename) if zipinfo & 0x800
+              len(ascii encoding of filename) otherwise
+      * [5] - (zipinfo.date_time[0] - 1980) << 9 |
+               zipinfo.date_time[1] << 5 | zipinfo.date_time[2]
+      * [6] - (zipinfo.date_time[3] - 1980) << 11 |
+               zipinfo.date_time[4] << 5 | (zipinfo.date_time[5] // 2)
+      * [7] - zipinfo.CRC
+    """
+    zipinfo = dict()
+    zfile = zipfile.ZipFile(path)
+    #Got ZipFile has not __exit__ on python 3.1
+    try:
+        for zitem in zfile.namelist():
+            zpath = zitem.replace('/', os.sep)
+            zipinfo[zpath] = zfile.getinfo(zitem)
+            assert zipinfo[zpath] is not None
+    finally:
+        zfile.close()
+    return zipinfo
 
 
 class ZipProvider(EggProvider):
@@ -1358,7 +1415,7 @@ class ZipProvider(EggProvider):
 
     def __init__(self, module):
         EggProvider.__init__(self,module)
-        self.zipinfo = zipimport._zip_directory_cache[self.loader.archive]
+        self.zipinfo = build_zipmanifest(self.loader.archive)
         self.zip_pre = self.loader.archive+os.sep
 
     def _zipinfo_name(self, fspath):
@@ -1392,6 +1449,14 @@ class ZipProvider(EggProvider):
                 self._extract_resource(manager, self._eager_to_zip(name))
         return self._extract_resource(manager, zip_path)
 
+    @staticmethod
+    def _get_date_and_size(zip_stat):
+        size = zip_stat.file_size
+        date_time = zip_stat.date_time + (0, 0, -1)  # ymdhms+wday, yday, dst
+        #1980 offset already done
+        timestamp = time.mktime(date_time)
+        return timestamp, size
+
     def _extract_resource(self, manager, zip_path):
 
         if zip_path in self._index():
@@ -1401,28 +1466,19 @@ class ZipProvider(EggProvider):
                 )
             return os.path.dirname(last)  # return the extracted directory name
 
-        zip_stat = self.zipinfo[zip_path]
-        t,d,size = zip_stat[5], zip_stat[6], zip_stat[3]
-        date_time = (
-            (d>>9)+1980, (d>>5)&0xF, d&0x1F,                      # ymd
-            (t&0xFFFF)>>11, (t>>5)&0x3F, (t&0x1F) * 2, 0, 0, -1   # hms, etc.
-        )
-        timestamp = time.mktime(date_time)
+        timestamp, size = self._get_date_and_size(self.zipinfo[zip_path])
 
+        if not WRITE_SUPPORT:
+            raise IOError('"os.rename" and "os.unlink" are not supported '
+                          'on this platform')
         try:
-            if not WRITE_SUPPORT:
-                raise IOError('"os.rename" and "os.unlink" are not supported '
-                              'on this platform')
 
             real_path = manager.get_cache_path(
                 self.egg_name, self._parts(zip_path)
             )
 
-            if os.path.isfile(real_path):
-                stat = os.stat(real_path)
-                if stat.st_size==size and stat.st_mtime==timestamp:
-                    # size and stamp match, don't bother extracting
-                    return real_path
+            if self._is_current(real_path, zip_path):
+                return real_path
 
             outf, tmpnam = _mkstemp(".$extract", dir=os.path.dirname(real_path))
             os.write(outf, self.loader.get_data(zip_path))
@@ -1435,11 +1491,9 @@ class ZipProvider(EggProvider):
 
             except os.error:
                 if os.path.isfile(real_path):
-                    stat = os.stat(real_path)
-
-                    if stat.st_size==size and stat.st_mtime==timestamp:
-                        # size and stamp match, somebody did it just ahead of
-                        # us, so we're done
+                    if self._is_current(real_path, zip_path):
+                        # the file became current since it was checked above,
+                        #  so proceed.
                         return real_path
                     elif os.name=='nt':     # Windows, del old file and retry
                         unlink(real_path)
@@ -1452,6 +1506,23 @@ class ZipProvider(EggProvider):
 
         return real_path
 
+    def _is_current(self, file_path, zip_path):
+        """
+        Return True if the file_path is current for this zip_path
+        """
+        timestamp, size = self._get_date_and_size(self.zipinfo[zip_path])
+        if not os.path.isfile(file_path):
+            return False
+        stat = os.stat(file_path)
+        if stat.st_size!=size or stat.st_mtime!=timestamp:
+            return False
+        # check that the contents match
+        zip_contents = self.loader.get_data(zip_path)
+        f = open(file_path, 'rb')
+        file_contents = f.read()
+        f.close()
+        return zip_contents == file_contents
+
     def _get_eager_resources(self):
         if self.eagers is None:
             eagers = []
@@ -1594,7 +1665,7 @@ class EggMetadata(ZipProvider):
     def __init__(self, importer):
         """Create a metadata provider from a zipimporter"""
 
-        self.zipinfo = zipimport._zip_directory_cache[importer.archive]
+        self.zipinfo = build_zipmanifest(importer.archive)
         self.zip_pre = importer.archive+os.sep
         self.loader = importer
         if importer.prefix:
@@ -2825,3 +2896,4 @@ run_main = run_script   # backward compatibility
 add_activation_listener(lambda dist: dist.activate())
 working_set.entries=[]; map(working_set.add_entry,sys.path) # match order
 
+
